using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Data;
using System.Drawing;
using System.Text;
using System.Windows.Forms;
using System.Xml;
using System.IO;
using System.Net;
using System.Net.Sockets;

/*
 *  ***********************
 *  *      RarMount2      *
 *  ***********************
 * 
 *  Author JackJson
 * 
 *   Changelog:
 *   2008-02-03  First release
 * 
 *  If you change the source code, please notify me.
 * 
 *  http://forum.team-mediaportal.com/streaming_rar_files-t12448.html
 * 
 */

/*
 * USAGE, mount:
 *  rarmount2 "c:\folder\archive.rar"
 *  imdisk -a -t proxy -m M: -u 5 -o ro,rem,ip -f 127.0.0.1:32123 -b 0 -s 31G
 * Unmount:
 *  imdisk -d -u 5
 * 
 * Additional information:
 * ImDisk:  http://www.ltr-data.se/opencode.html#ImDisk
 * Old RAR: http://www.bsdg.org/SWAG/DATATYPE/0035.PAS.html
 * New RAR: http://darkshire.org/~jhkim/src/rar/technote.txt
 *
 * Abilities:
 *  Mount uncompressed RAR file and stream it's content, for example an AVI
 *
 * Limits:
 *  The RAR archive must be uncompressed (i.e. stored mode)
 *  The RAR archive may only contain one file (fairly easy to fix)
 *  Limited NTFS support, files larger than ~15 GiB are most likely not supported
 *  Requires ImDisk (since I'm not skilled enought to write my own drivers)
 *  Etc, etc, etc
 * 
 * Next release will hopefully:
 *  Find SRT/SUB file in RAR dir and mount it beside the uncompressed RAR content
 */

namespace RarMount2
{

    public partial class frmMain : Form
    {
        private bool boolDebugMode = true;
        public static string[] Args = null;
        private bool boolSilentMode = false;
        private uint intListenPort;
        private string strMainFile = "";
        private string strNtfsHelpFile = "";
        private SimpleNTFS ntfsFileSystem;

        public frmMain()
        {
            InitializeComponent();
        }

        private void frmMain_Load(object sender, EventArgs e)
        {
            // Check debug mode setting
            string strDebugMode = getSetting("DEBUG");
            if (!strDebugMode.Equals("True") && !strDebugMode.Equals(""))
                boolDebugMode = false;

            debugPrint("Application started");

            // Read settings from file
            uint.TryParse(getSetting("LISTENPORT"), out intListenPort);
            if (intListenPort == 0)
                intListenPort = 32123; // Default listen port

            // Get NTFS help file
            strNtfsHelpFile = getSetting("NTFSFILE");
            if (strNtfsHelpFile.Length == 0)
                strNtfsHelpFile = "ntfs.bin"; // Default NTFS help file

            // Check command line arguments
            for (int i = 0; i < Args.Length; i++)
            {
                if (Args[i].ToLower() == "/silent")
                {
                    boolSilentMode = true;
                }
                else
                {
                    // Presume RAR archive
                    strMainFile = Args[i];
                }
            }

            // If not silent mode, show form
            if (!boolSilentMode)
            {
                this.WindowState = FormWindowState.Normal;
                this.ShowInTaskbar = true;
            }
            else
            {
                this.ShowInTaskbar = true;
            }
            
        }

        // Reads setting from XML file
        private string getSetting(string strSettingTitle)
        {
            XmlDocument xmlSettingsDoc = new XmlDocument();
            try
            {
                // Open and read settings file
                xmlSettingsDoc.Load("rarmount_settings.xml");
                return xmlSettingsDoc.GetElementsByTagName(strSettingTitle).Item(0).InnerText.ToString();
            }
            catch (Exception e)
            {
                // Error while reading settings
                debugPrint("Error while reading setting '" + strSettingTitle  + "': " + e.Message.ToString());
                return "";
            }
        }

        // Saves a setting
        private void saveSettings()
        {
            XmlDocument xmlSettingsDoc = new XmlDocument();
            XmlElement xmlSettingsElement = xmlSettingsDoc.CreateElement("SETTINGS");
            XmlElement xmlValueElement;
            string[] strNames  = new string[] { "DEBUG", "LISTENPORT", "NTFSFILE" };
            string[] strValues = new string[] { boolDebugMode.ToString(), intListenPort.ToString(), strNtfsHelpFile };
            for (int i = 0; i < strNames.Length; i++)
            {
                xmlValueElement = xmlSettingsDoc.CreateElement(strNames[i]);
                xmlValueElement.InnerText = strValues[i];
                xmlSettingsElement.AppendChild(xmlValueElement);
            }
            xmlSettingsDoc.AppendChild(xmlSettingsElement);
            xmlSettingsDoc.Save("rarmount_settings.xml");
            debugPrint("Settings saved");
        }

        private void debugPrint(string strDebugMessage)
        {
            // If debug is enabled ...
            if (boolDebugMode)
            {
                // ... write to debug file
                StreamWriter swFile = File.AppendText("rarmount_debug.txt");
                swFile.Write(DateTime.Now.ToString());
                swFile.Write(" - ");
                swFile.WriteLine(strDebugMessage);
                swFile.Close();
            }
        }

        private void frmMain_Shown(object sender, EventArgs e)
        {
            try
            {
                FileInfo fiMain = new FileInfo(strMainFile);
                if (!fiMain.Exists)
                {
                    setStatus("Archive not found!");
                    debugPrint("Archive not found: " + strMainFile);
                    exitWithDelay();
                }
                else
                {
                    RarCollection rcMain = new RarCollection(strMainFile);
                    if (rcMain.filesLoadedOk())
                    {
                        setStatus(rcMain.getLoadedFilesCount() + " RAR file(s) loaded");
                        debugPrint(rcMain.getLoadedFilesCount() + " RAR file(s) loaded successfully");

                        // Init NTFS
                        if (initNTFSfs(rcMain))
                        {
                            runImDiskListner();
                        }
                        else
                        {
                            setStatus("Error while initiating NTFS");
                            debugPrint("Error while initiating simple NTFS file system, exiting");
                        }

                        exitWithDelay();
                    }
                    else
                    {
                        setStatus("Error while loading RAR file(s)");
                        debugPrint("Error while loading RAR file(s)");
                        exitWithDelay();
                    }
                }
            }
            catch (Exception err)
            {
                setStatus("Error occured!");
                debugPrint("Error occured when mounting " + strMainFile + ": " + err.Message);
                exitWithDelay();
            }
        }

        // Initiate NTFS filesystem
        private bool initNTFSfs(RarCollection rcMain)
        {
            FileInfo fiNtfsFile = new FileInfo(strNtfsHelpFile);

            // Does the help file (default: ntfs.bin) even exist?
            if (!fiNtfsFile.Exists)
            {
                debugPrint("NTFS binary help file not found: " + strNtfsHelpFile);
                return false;
            }
            else if (fiNtfsFile.Length != 2611200)
            {
                debugPrint("NTFS binary help file has wrong file size, expected 2611200, got: " + fiNtfsFile.Length);
                return false;
            }

            ntfsFileSystem = new SimpleNTFS(fiNtfsFile.FullName, rcMain);

            if (ntfsFileSystem.loadedOk())
            {
                debugPrint("NTFS file system initiated successfully");
                return true;
            }
            else
            {
                return false;
            }
        }

        // Start and run ImDisk listner
        private void runImDiskListner()
        {
            TcpListener tlListener = new TcpListener(IPAddress.Any, (int)intListenPort);
            tlListener.Start();

            setStatus("Waiting for connection ...");
            debugPrint("Listning for connection on port " + intListenPort);
            while (!tlListener.Pending())
            {
                // If form is closed, exit listner
                if (this.IsDisposed)
                    break;
                Application.DoEvents();
            }
            if (tlListener.Pending())
            {
                TcpClient tcClient = tlListener.AcceptTcpClient();
                setStatus("Connected to: " + tcClient.Client.RemoteEndPoint.ToString());
                debugPrint("Accepted request from " + tcClient.Client.RemoteEndPoint.ToString());
                if (tcClient.Connected)
                {
                    NetworkStream nsStream = tcClient.GetStream();

                    nsStream.ReadTimeout = 1000;

                    byte[] byteSendBuffer;
                    byte[] byteFileReadBuffer;

                    byte[] byteReadBuffer = new byte[1024];
                    int intReceivedDataLength = -1; // <- start value
                    while (intReceivedDataLength != 0)
                    {
                        try
                        {
                            // Ugly (but working) way of determine if the connection is lost.
                            // If read timeout occured, the connection is still alive. 
                            // If 0 byte is returned, but no error, the connection is lost.
                            try
                            {
                                intReceivedDataLength = nsStream.Read(byteReadBuffer, 0, 1024);

                                if (intReceivedDataLength > 0)
                                {
                                    ulong lngResponseCode = BitConverter.ToUInt64(byteReadBuffer, 0);
                                    switch (lngResponseCode)
                                    {
                                        case 0:
                                            //debugPrint("Command: 0 (IMDPROXY_NULL_REQ)");
                                            break;
                                        case 1:
                                            setStatus("Connected to ImDisk");
                                            //debugPrint("Command: 1 (IMDPROXY_INFO_REQ)");
                                            ulong disk_size = 33285996544; // 31 GiB
                                            ulong req_alignment = 512;
                                            ulong flags = 0x1; //read only
                                            byteSendBuffer = new byte[sizeof(ulong) * 3];
                                            BitConverter.GetBytes(disk_size).CopyTo(byteSendBuffer, sizeof(ulong) * 0);
                                            BitConverter.GetBytes(req_alignment).CopyTo(byteSendBuffer, sizeof(ulong) * 1);
                                            BitConverter.GetBytes(flags).CopyTo(byteSendBuffer, sizeof(ulong) * 2);
                                            //debugPrint("Sending IMDPROXY_INFO_RESP");
                                            nsStream.Write(byteSendBuffer, 0, byteSendBuffer.Length);
                                            break;
                                        case 2:
                                            ulong read_offset = BitConverter.ToUInt64(byteReadBuffer, sizeof(ulong) * 1);
                                            ulong read_length = BitConverter.ToUInt64(byteReadBuffer, sizeof(ulong) * 2);
                                            //debugPrint("Command: 2 (IMDPROXY_READ_REQ), offset: " + read_offset.ToString() + ", length: " + read_length.ToString());
                                            byteFileReadBuffer = new byte[(int)read_length];
                                            int intFileReadBytes = ntfsFileSystem.readData(byteFileReadBuffer, read_offset, (int)read_length);
                                            ulong read_errorno = 0; // =9 if some problem occured
                                            ulong read_resp_length = (ulong)intFileReadBytes;
                                            byteSendBuffer = new byte[sizeof(ulong) * 2];
                                            BitConverter.GetBytes(read_errorno).CopyTo(byteSendBuffer, sizeof(ulong) * 0);
                                            BitConverter.GetBytes(read_resp_length).CopyTo(byteSendBuffer, sizeof(ulong) * 1);
                                            //debugPrint("Sending IMDPROXY_READ_RESP");
                                            nsStream.Write(byteSendBuffer, 0, byteSendBuffer.Length); // ERROR!
                                            nsStream.Write(byteFileReadBuffer, 0, intFileReadBytes);
                                            break;
                                        case 3:
                                            //debugPrint("Command: 3 (IMDPROXY_WRITE_REQ)");
                                            // Info: Errorno 9 = EBADF
                                            ulong write_errorno = 9;
                                            ulong write_length = 0;
                                            byteSendBuffer = new byte[sizeof(ulong) * 2];
                                            BitConverter.GetBytes(write_errorno).CopyTo(byteSendBuffer, sizeof(ulong) * 0);
                                            BitConverter.GetBytes(write_length).CopyTo(byteSendBuffer, sizeof(ulong) * 1);
                                            //debugPrint("Sending IMDPROXY_WRITE_RESP");
                                            nsStream.Write(byteSendBuffer, 0, byteSendBuffer.Length);
                                            break;
                                        case 4:
                                            //debugPrint("Command: 4 (IMDPROXY_CONNECT_REQ)");
                                            ulong connect_errorno = 0;
                                            ulong connect_object_ptr = 0;
                                            byteSendBuffer = new byte[sizeof(ulong) * 2];
                                            BitConverter.GetBytes(connect_errorno).CopyTo(byteSendBuffer, sizeof(ulong) * 0);
                                            BitConverter.GetBytes(connect_object_ptr).CopyTo(byteSendBuffer, sizeof(ulong) * 1);
                                            //debugPrint("Sending IMDPROXY_CONNECT_RESP");
                                            nsStream.Write(byteSendBuffer, 0, byteSendBuffer.Length);
                                            break;
                                        default:
                                            debugPrint("Unknow ImDisk command: " + lngResponseCode.ToString());
                                            break;
                                    }

                                }

                            }
                            catch
                            {
                                Application.DoEvents();
                            }
                            
                        }
                        catch (Exception err)
                        {
                            debugPrint("ImDiskListner error, " + err.Message);
                            intReceivedDataLength = 0;
                        }
                    }
                    nsStream.Close();
                }
                tcClient.Close();
            }
            tlListener.Stop();

            setStatus("Connection lost, exiting ...");
            debugPrint("Connection to ImDisk lost");
        }

        // Set main window status label
        private void setStatus(string strStatusMessage)
        {
            lblStatus.Text = strStatusMessage;
            this.Refresh();
        }

        private void exitWithDelay()
        {
            // Sleep for 3 seconds (30 * 100 ms)
            // New 2008-05-24: 1 sec is enough
            for (int i = 0; i < 10; i++)
            {
                System.Threading.Thread.Sleep(100);
                Application.DoEvents();
            }
            this.Close();
        }

        private void frmMain_FormClosed(object sender, FormClosedEventArgs e)
        {
            saveSettings();
            debugPrint("Application closed");
        }

    }

    class RarCollection
    {
        private RarFile[] RarFiles;
        private string strRarFileName;   // <- name of stored file
        private long lngRarFileFullSize; // <- full file size of stored file
        private bool boolFilesLoadedOk = false; // <- indicates the status

        public RarCollection(string strMainFileName)
        {
            boolFilesLoadedOk = findArchives(strMainFileName); // find spanned files and load info
        }

        // N.B! Offset is reading offset, not returning offset.
        public int readData(byte[] byteRarReadBuffer, ulong lngReadOffset, int intReadLength)
        {
            int intByteFilled = 0;
            int intReadMax = 0;
            long lngCurrPosition;
            foreach (RarFile rfCurrRar in RarFiles)
            {
                if (lngReadOffset >= rfCurrRar.getRarFileOffset() && lngReadOffset < rfCurrRar.getRarFileOffset() + (ulong)rfCurrRar.getRarFileSize())
                {
                    // Requested data starts within this file
                    lngCurrPosition = (long)lngReadOffset - (long)rfCurrRar.getRarFileOffset() + rfCurrRar.getRarFileStart();
                    if (lngCurrPosition - rfCurrRar.getRarFileStart() + (long)intReadLength > rfCurrRar.getRarFileSize())
                        intReadMax = (int)(rfCurrRar.getRarFileSize() - lngCurrPosition + rfCurrRar.getRarFileStart());
                    else
                        intReadMax = intReadLength;
                    intByteFilled = rfCurrRar.readData(lngCurrPosition, byteRarReadBuffer, intByteFilled, intReadMax);
                }
                else if (intByteFilled > 0 && intByteFilled < intReadLength)
                {
                    // Last file didn't contain all requested data
                    lngCurrPosition = rfCurrRar.getRarFileStart();
                    if (intReadLength - intByteFilled > rfCurrRar.getRarFileSize())
                        intReadMax = (int)rfCurrRar.getRarFileSize();
                    else
                        intReadMax = intReadLength - intByteFilled;
                    intByteFilled = intByteFilled + rfCurrRar.readData(lngCurrPosition, byteRarReadBuffer, intByteFilled, intReadMax);
                }
                else if (intByteFilled == intReadLength)
                {
                    // All data is here :)
                    return intByteFilled;
                }
            }
            // Didn't get all data. Return it anyway.
            return intByteFilled;
        }

        public int getLoadedFilesCount()
        {
            return RarFiles.Length;
        }

        public bool filesLoadedOk()
        {
            return boolFilesLoadedOk;
        }

        public string getStoredFileName()
        {
            return strRarFileName;
        }

        public long getStoredFileSize()
        {
            return lngRarFileFullSize;
        }

        private bool findArchives(string strStartingFile)
        {
            FileInfo fiMain = new FileInfo(strStartingFile);
            string strFileSearchPattern = "";
            if ((fiMain.Name.Length > 12 && fiMain.Name.ToLower().Substring(fiMain.Name.Length - 12, 12) == ".part001.rar") ||
                (fiMain.Name.Length > 11 && fiMain.Name.ToLower().Substring(fiMain.Name.Length - 11, 11) == ".part01.rar")) // Fixed bug 2008-05-24 (found with Kyle.XY.S02E17.REAL.REPACK.DSR.XviD-0TV)
            {
                strFileSearchPattern = fiMain.Name.Substring(0, fiMain.Name.Length - 7) + "*.rar";
            }
            else if (fiMain.Name.ToLower().Substring(fiMain.Name.Length - 4, 4) == ".rar")
            {
                strFileSearchPattern = fiMain.Name.Substring(0, fiMain.Name.Length - 4) + ".r*";
            }

            
            FileInfo[] fiFiles = fiMain.Directory.GetFiles(strFileSearchPattern);
            RarFiles = new RarFile[fiFiles.Length];
            RarFiles[0] = new RarFile(fiMain.FullName, 0);
            if (!RarFiles[0].loadFileInfo())
                return false; // <- error occured in loadFileInfo()

            strRarFileName = RarFiles[0].getRarFileName();
            lngRarFileFullSize = RarFiles[0].getRarFileFullSize();

            int intRarNo = 1;
            foreach (FileInfo fiPart in fiFiles)
            {
                if (fiPart.Name != fiMain.Name)
                {
                    RarFiles[intRarNo] = new RarFile(fiPart.FullName, (long)RarFiles[intRarNo - 1].getRarFileOffset() + RarFiles[intRarNo - 1].getRarFileSize());
                    if (!RarFiles[intRarNo].loadFileInfo())
                        return false; // <- error occured in loadFileInfo()
                    intRarNo++;
                }
            }

            return true;
        }
    }

    class RarFile
    {
        private string strFileName;  // <- full path to file
        private long lngFileSize;    // <- rar file size
        private FileStream fsReader;

        private long lngRarFileSize;    // <- size of stored file contained in this rar file
        private long lngRarFileStart;   // <- offset to where the real file data begins (i.e. rar header + previous files)
        private ulong lngRarFileOffset; // <- offset in relation to lngRarFileFullSize (ex: if .rar contains the first 100byte of myfile.txt, .r01 will have offset 100)

        // Following variables are only important in first file (.rar, .part001.rar)
        private string strRarFileName;   // <- name of stored file
        private long lngRarFileFullSize; // <- full file size of stored file

        public RarFile(string strLoadFileName, long lngLoadRarFileOffset)
        {
            strFileName = strLoadFileName;
            lngRarFileOffset = (ulong)lngLoadRarFileOffset;
            FileInfo fiFile = new FileInfo(strLoadFileName);
            lngFileSize = fiFile.Length;
        }

        public int readData(long lngPosition, byte[] byteReadBuffer, int intOffset, int intCount)
        {
            fsReader.Position = lngPosition;
            return fsReader.Read(byteReadBuffer, intOffset, intCount);
        }

        public bool loadFileInfo()
        {
            // Open RAR file for reading
            try
            {
                fsReader = new FileStream(strFileName, FileMode.Open, FileAccess.Read);

                // Check RAR file header
                if (!checkFileHeader())
                {
                    // Wrong RAR file header!
                    return false;
                }

                // Read rar archive header size, 2 byte (16 bit) int 
                //  starting at position 12 (7 byte file header + 5 byte unneeded data)
                int intRarHeaderSize = readInt16(12);

                // Read file header flags to determine if compressed file is larger than 2gb
                bool boolLargeFileMode;
                byte[] byteRarFileSize = new byte[8];
                byte[] byteRarFileFullSize = new byte[8];
                int intFileHeaderFlags = readInt8(11 + (long)intRarHeaderSize);
                string strFileHeaderFlags = Convert.ToString(intFileHeaderFlags, 2);
                if (strFileHeaderFlags.Substring(strFileHeaderFlags.Length - 1, 1) == "1")
                    boolLargeFileMode = true; // 64-bit file size
                else
                    boolLargeFileMode = false; // 32-bit file size

                // Read rar archive data starting position (where the stored file begins)
                //lngRarFileStart = (long)readInt16(12 + (long)intRarHeaderSize) + (long)intRarHeaderSize + 7;
                lngRarFileStart = (long)readInt16(-1) + (long)intRarHeaderSize + 7;

                // Read stored file data size (no of bytes stored in current archive, 
                //  i.e. around 15mb for a spanned xvid movie)
                readByte32(-1).CopyTo(byteRarFileSize, 0);

                // Read stored file full size (if extracted, i.e. 700mb for a xvid movie)
                readByte32(-1).CopyTo(byteRarFileFullSize, 0);

                // Read type of compression
                if (readInt8(fsReader.Position + 10) != 0x30) // Skip some unneeded data, then read 1 byte
                    return false; // <- archive is compressed !

                // Read stored file name size
                int intRarFileNameLength = readInt16(-1);

                fsReader.Position = fsReader.Position + 4; // Skip unneeded data

                // If large file mode (> 2gb), read high 4 bytes of 64-bit file sizes
                if (boolLargeFileMode)
                {
                    readByte32(-1).CopyTo(byteRarFileSize, 4);
                    readByte32(-1).CopyTo(byteRarFileFullSize, 4);
                }
                lngRarFileSize = BitConverter.ToInt64(byteRarFileSize, 0);
                lngRarFileFullSize = BitConverter.ToInt64(byteRarFileFullSize, 0);

                // Read stored file name
                byte[] byteRarReadBuffer = new byte[intRarFileNameLength];
                fsReader.Read(byteRarReadBuffer, 0, byteRarReadBuffer.Length);
                strRarFileName = Encoding.ASCII.GetString(byteRarReadBuffer);

                return true; // All ok!
            }
            catch
            {
                return false;
            }
        }

        private int readInt8(long lngPosition)
        {
            if (lngPosition >= 0)
                fsReader.Position = lngPosition;
            return fsReader.ReadByte();
        }

        private int readInt16(long lngPosition)
        {
            if (lngPosition >= 0)
                fsReader.Position = lngPosition;
            byte[] byteRarReadBuffer = new byte[2];
            fsReader.Read(byteRarReadBuffer, 0, byteRarReadBuffer.Length);
            return (int)BitConverter.ToInt16(byteRarReadBuffer, 0);
        }

        private byte[] readByte32(long lngPosition)
        {
            if (lngPosition >= 0)
                fsReader.Position = lngPosition;
            byte[] byteRarReadBuffer = new byte[4];
            fsReader.Read(byteRarReadBuffer, 0, byteRarReadBuffer.Length);
            return byteRarReadBuffer;
        }

        private long readInt32(long lngPosition)
        {
            if (lngPosition >= 0)
                fsReader.Position = lngPosition;
            byte[] byteRarReadBuffer = new byte[4];
            fsReader.Read(byteRarReadBuffer, 0, byteRarReadBuffer.Length);
            return (long)BitConverter.ToInt32(byteRarReadBuffer, 0);
        }

        public ulong getRarFileOffset()
        {
            return lngRarFileOffset;
        }

        public long getRarFileSize()
        {
            return lngRarFileSize;
        }

        public long getRarFileStart()
        {
            return lngRarFileStart;
        }

        public string getRarFileName()
        {
            return strRarFileName;
        }

        public long getRarFileFullSize()
        {
            return lngRarFileFullSize;
        }

        // Look for valid RAR file header
        private bool checkFileHeader()
        {
            byte[] byteRealRarHeader = new byte[7] { 0x52, 0x61, 0x72, 0x21, 0x1A, 0x07, 0x00 };
            byte[] byteFileHeader = new byte[7];
            fsReader.Position = 0;
            fsReader.Read(byteFileHeader, 0, 7);
            bool boolFileHeaderIsRar = true;
            for (int i = 0; i < 7; i++)
            {
                if (byteFileHeader[i] != byteRealRarHeader[i])
                {
                    boolFileHeaderIsRar = false;
                    break;
                }
            }
            return boolFileHeaderIsRar;
        }
    }

    // Simulates a 31 GiB NTFS harddrive, based on reverse engineering
    /* INFO
     * A 31GiB NTFS partition with one file, 15 GiB UNPACKED.RAW, looks something like this:
     * Byte from    Byte to     Description
     * 0            1476095     NTFS header, device info, etc.
     * 22446080     22450175    INDX record, dictonary info?
     * 3154083840   3154710527  RSTR record and some (unused?) file info
     * 3154710528   3221192703  Probably allocation info, in this case every byte is hex: FF
     * 3221192704   3221270527  File info again, incl file name and file size
     * 16642965504  16643391487 File info again, incl file name and file size
     * 16643391488  32749518847 File UNPACKED.RAW 15 GiB data
     * 33285996032  33285996543 NTFS footer
     * Everywhere else          Every byte is hex: 00
     */
    class SimpleNTFS
    {
        private string strBinaryHelpFile = "";
        private byte[] byte15gbFileType;
        private long lng15gbFileSize = 0;
        RarCollection rcMain;
        private FileStream fsHelpFileReader;

        /* FIXME: Future support for SRT/SUB
        public SimpleNTFS(string strHelpFile, string strUncompressedFile, string strSubtitleFile)
        {
        }*/

        public SimpleNTFS(string strHelpFile, RarCollection rarUncompressedFile)
        {
            strBinaryHelpFile = strHelpFile;
            rcMain = rarUncompressedFile;
            if (rcMain.filesLoadedOk())
            {
                try
                {
                    fsHelpFileReader = new FileStream(strBinaryHelpFile, FileMode.Open, FileAccess.Read);
                    string str15gbFileName = rcMain.getStoredFileName();
                    string str15gbFileType = str15gbFileName.Substring(str15gbFileName.Length - 3, 3).ToUpper();
                    byte15gbFileType = Encoding.ASCII.GetBytes(str15gbFileType);
                    lng15gbFileSize = rcMain.getStoredFileSize();
                }
                catch
                {
                }
            }
        }

        public bool loadedOk()
        {
            // We should do some more checks here ...
            try
            {
                if (lng15gbFileSize > 0 && fsHelpFileReader.Length > 0)
                    return true;
                else
                    return false;
            }
            catch
            {
                return false;
            }
        }

        public int readData(byte[] byteNtfsReadBuffer, ulong lngNtfsReadOffset, int intReadLength)
        {
            // FIXME:
            //  Precache byte arrays with file size and name (no need to regenerate every time accessed)
            //  Comments!
            
            int intReadMax;
            int intBytesLeftToRead;
            int intBytesReadTotal = 0;
            ulong lngReadOffset;
            while (intReadLength > intBytesReadTotal)
            {
                intBytesLeftToRead = intReadLength - intBytesReadTotal;
                lngReadOffset = lngNtfsReadOffset + (ulong)intBytesReadTotal;
                if (lngReadOffset >= 16643391488 && lngReadOffset < 16643391488 + (ulong)lng15gbFileSize)
                {
                    if (lngReadOffset + (ulong)intBytesLeftToRead > 16643391488 + (ulong)lng15gbFileSize)
                        intReadMax = (int)((16643391488 + (ulong)lng15gbFileSize) - lngReadOffset);
                    else
                        intReadMax = intBytesLeftToRead;
                    ulong lngStartPosition = lngReadOffset - 16643391488;
                    int intBytesRead = 0;
                    intBytesRead = rcMain.readData(byteNtfsReadBuffer, lngStartPosition, intReadMax);
                    // Fixed bug 2008-05-24, avoid never ending loop
                    if (intBytesRead == 0)
                        break;
                    // /Fixed bug

                    intBytesReadTotal += intBytesRead;
                }
                else if (lngReadOffset >= 0 && lngReadOffset < 1476096)
                {
                    intBytesReadTotal = ntfsHelpFileRead(0, 0, 1476096, byteNtfsReadBuffer, lngReadOffset, intBytesReadTotal, intBytesLeftToRead);
                }
                else if (lngReadOffset >= 22446080 && lngReadOffset < 22450176)
                {
                    intBytesReadTotal = ntfsHelpFileRead(1476096, 22446080, 22450176, byteNtfsReadBuffer, lngReadOffset, intBytesReadTotal, intBytesLeftToRead);
                }
                else if (lngReadOffset >= 3154083840 && lngReadOffset < 3154710528)
                {
                    intBytesReadTotal = ntfsHelpFileRead(1480192, 3154083840, 3154710528, byteNtfsReadBuffer, lngReadOffset, intBytesReadTotal, intBytesLeftToRead);
                }
                else if (lngReadOffset >= 3221192704 && lngReadOffset < 3221269764) //(lngReadOffset >= 3221192704 && lngReadOffset < 3221270528)
                {
                    intBytesReadTotal = ntfsHelpFileRead(2106880, 3221192704, 3221269764, byteNtfsReadBuffer, lngReadOffset, intBytesReadTotal, intBytesLeftToRead);
                }
                else if (lngReadOffset >= 3221269764 && lngReadOffset < 3221269840)
                {
                    // File name and size (x2)
                    byte[] byteFileSizeData = new byte[76];
                    BitConverter.GetBytes(lng15gbFileSize).CopyTo(byteFileSizeData, 60);
                    BitConverter.GetBytes(lng15gbFileSize).CopyTo(byteFileSizeData, 68);
                    byteFileSizeData[0] = byte15gbFileType[0]; // A
                    byteFileSizeData[2] = byte15gbFileType[1]; // V
                    byteFileSizeData[4] = byte15gbFileType[2]; // I
                    byteFileSizeData[12] = 0x80;
                    byteFileSizeData[16] = 0x48;
                    byteFileSizeData[20] = 0x01;
                    byteFileSizeData[26] = 0x01;
                    byteFileSizeData[36] = 0xFF;
                    byteFileSizeData[37] = 0x7F;
                    byteFileSizeData[38] = 0x07;
                    byteFileSizeData[44] = 0x40;
                    byteFileSizeData[55] = 0xC0;
                    byteFileSizeData[56] = 0x03;


                    if (lngReadOffset + (ulong)intBytesLeftToRead > 3221269840)
                        intReadMax = (int)(3221269840 - lngReadOffset);
                    else
                        intReadMax = intBytesLeftToRead;

                    int intStartPosition = (int)(lngReadOffset - 3221269764);
                    for (int i = 0; i < intReadMax; i++)
                    {
                        byteNtfsReadBuffer[i + intBytesReadTotal] = byteFileSizeData[i + intStartPosition];
                    }
                    intBytesReadTotal += intReadMax;
                }
                else if (lngReadOffset >= 3221269840 && lngReadOffset < 3221270528)
                {
                    intBytesReadTotal = ntfsHelpFileRead(2184016, 3221269840, 3221270528, byteNtfsReadBuffer, lngReadOffset, intBytesReadTotal, intBytesLeftToRead);
                }
                else if (lngReadOffset >= 16642965504 && lngReadOffset < 16643065136) //(lngReadOffset >= 16642965504 && lngReadOffset < 16643391488)
                {
                    intBytesReadTotal = ntfsHelpFileRead(2184704, 16642965504, 16643065136, byteNtfsReadBuffer, lngReadOffset, intBytesReadTotal, intBytesLeftToRead);
                }
                else if (lngReadOffset >= 16643065136 && lngReadOffset < 16643065177)
                {
                    // File name and size
                    byte[] byteFileSizeData = new byte[41];
                    BitConverter.GetBytes(lng15gbFileSize).CopyTo(byteFileSizeData, 0);
                    byteFileSizeData[8] = 0x20;
                    byteFileSizeData[16] = 0x0C;
                    byteFileSizeData[17] = 0x03;
                    byteFileSizeData[18] = 0x55;
                    byteFileSizeData[20] = 0x4E;
                    byteFileSizeData[22] = 0x50;
                    byteFileSizeData[24] = 0x41;
                    byteFileSizeData[26] = 0x43;
                    byteFileSizeData[28] = 0x4B;
                    byteFileSizeData[30] = 0x45;
                    byteFileSizeData[32] = 0x44;
                    byteFileSizeData[34] = 0x2E;
                    byteFileSizeData[36] = byte15gbFileType[0]; // A
                    byteFileSizeData[38] = byte15gbFileType[1]; // V
                    byteFileSizeData[40] = byte15gbFileType[2]; // I


                    if (lngReadOffset + (ulong)intBytesLeftToRead > 16643065177)
                        intReadMax = (int)(16643065177 - lngReadOffset);
                    else
                        intReadMax = intBytesLeftToRead;

                    int intStartPosition = (int)(lngReadOffset - 16643065136);
                    for (int i = 0; i < intReadMax; i++)
                    {
                        byteNtfsReadBuffer[i + intBytesReadTotal] = byteFileSizeData[i + intStartPosition];
                    }
                    intBytesReadTotal += intReadMax;
                }
                else if (lngReadOffset >= 16643065177 && lngReadOffset < 16643391488)
                {
                    intBytesReadTotal = ntfsHelpFileRead(2284377, 16643065177, 16643391488, byteNtfsReadBuffer, lngReadOffset, intBytesReadTotal, intBytesLeftToRead);
                }
                else if (lngReadOffset >= 33285996032 && lngReadOffset < 33285996544)
                {
                    intBytesReadTotal = ntfsHelpFileRead(2610688, 33285996032, 33285996544, byteNtfsReadBuffer, lngReadOffset, intBytesReadTotal, intBytesLeftToRead);
                    // Last part of file, no more data to read
                    break;
                }
                else if (lngReadOffset >= 3154710528 && lngReadOffset < 3221192704)
                {
                    // Return hex(FF) filled array
                    if (lngReadOffset + (ulong)intBytesLeftToRead > 3221192704)
                        intReadMax = (int)(3221192704 - lngReadOffset);
                    else
                        intReadMax = intBytesLeftToRead;
                    for (int i = intBytesReadTotal; i < intBytesReadTotal + intReadMax; i++)
                    {
                        byteNtfsReadBuffer[i] = 255;
                    }
                    intBytesReadTotal += intReadMax;
                }
                else
                {
                    // Return zero filled array
                    intReadMax = 1; // Slow but safest
                    for (int i = intBytesReadTotal; i < intBytesReadTotal + intReadMax; i++)
                    {
                        byteNtfsReadBuffer[i] = 0;
                    }
                    intBytesReadTotal += intReadMax;
                }
            }
            return intBytesReadTotal;
        }

        private int ntfsHelpFileRead(int intHelpFileOffset, ulong lngLowerOffset, ulong lngUpperOffset, byte[] byteNtfsReadBuffer, ulong lngReadOffset, int intBytesReadTotal, int intBytesLeftToRead)
        {
            int intReadMax;
            if (lngReadOffset + (ulong)intBytesLeftToRead > lngUpperOffset)
                intReadMax = (int)(lngUpperOffset - lngReadOffset);
            else
                intReadMax = intBytesLeftToRead;
            fsHelpFileReader.Position = (long)(lngReadOffset - lngLowerOffset) + (long)intHelpFileOffset;
            int intBytesRead = fsHelpFileReader.Read(byteNtfsReadBuffer, intBytesReadTotal, intReadMax);
            return intBytesReadTotal + intBytesRead;
        }
    }
}